import requests

class APIClient:
    """
    A class to make HTTP requests to the Compext AI API.
    """
    def __init__(self, base_url:str, api_key:str, timeout:int=10, retries:int=3):
        self.base_url = base_url + "/api/v1"
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}"
        }
        self.timeout = timeout
        self.retries = retries
    # set timeout for all the requests and retry if the request times out 
    def get(self, route:str, data:dict={},**kwargs):
        for _ in range(self.retries):
            try:
                response = requests.get(self.base_url + route, headers=self.headers, json=data, timeout=self.timeout, **kwargs)
                return {
                    "status": response.status_code,
                    "data": response.json()
                }
            except requests.exceptions.Timeout:
                continue
        return {
            "status": 500,
            "data": {}
        }
    
    def post(self, route:str, data:dict={},**kwargs):
        for _ in range(self.retries):
            try:
                response = requests.post(self.base_url + route, headers=self.headers, json=data,timeout=self.timeout, **kwargs)
                return {
                    "status": response.status_code,
                    "data": response.json()
                }
            except requests.exceptions.Timeout:
                continue
        return {
            "status": 500,
            "data": {}
        }
    
    def put(self, route:str, data:dict={},**kwargs):
        for _ in range(self.retries):
            try:
                response = requests.put(self.base_url + route, headers=self.headers, json=data,timeout=self.timeout, **kwargs)
                return {
                    "status": response.status_code,
                    "data": response.json()
                }
            except requests.exceptions.Timeout:
                continue
        return {
            "status": 500,
            "data": {}
        }

    def delete(self, route:str, data:dict={},**kwargs):
        for _ in range(self.retries):
            try:
                response = requests.delete(self.base_url + route, headers=self.headers, json=data,timeout=self.timeout, **kwargs)
                return {
                    "status": response.status_code,
                    "data": response.json()
                }
            except requests.exceptions.Timeout:
                continue
        return {
            "status": 500,
            "data": {}
        }




from pydantic import BaseModel, Field
from typing import Type

class HumanInTheLoopInputSchema(BaseModel):
    message: str = Field(description="The message to ask the human for input")

class HumanInTheLoop:
    name = "human_in_the_loop"
    description = "Use this tool to ask the human for input"
    input_schema: Type[BaseModel] = HumanInTheLoopInputSchema

    def _run(self, input:dict):
        pass



from compextAI.api.api import APIClient
from compextAI.messages import Message
from compextAI.threads import ThreadExecutionResponse
from compextAI.tools import get_tool
import time
import queue
import json
from compextAI.utils import *
from compextAI.tools import ToolResult

class ThreadExecutionStatus:
    status: str

    def __init__(self, status:str):
        self.status = status

def get_thread_execution_status(client:APIClient, thread_execution_id:str) -> str:
    response = client.get(f"/threadexec/{thread_execution_id}/status")

    status_code: int = response["status"]
    data: dict = response["data"]

    if status_code != 200:
        raise Exception(f"Failed to get thread execution status, status code: {status_code}, response: {data}")
    
    return ThreadExecutionStatus(data["status"])

def get_thread_execution_response(client:APIClient, thread_execution_id:str) -> dict:
    response = client.get(f"/threadexec/{thread_execution_id}/response")

    status_code: int = response["status"]
    data: dict = response["data"]

    if status_code != 200:
        raise Exception(f"Failed to get thread execution response, status code: {status_code}, response: {data}")
    
    return data


class ExecuteMessagesResponse:
    thread_execution_id: str

    def __init__(self, thread_execution_id:str, thread_execution_param_id:str, messages:list[Message], system_prompt:str, append_assistant_response:bool, metadata:dict):
        self.thread_execution_id = thread_execution_id
        self.thread_execution_param_id = thread_execution_param_id
        self.messages = messages
        self.system_prompt = system_prompt
        self.append_assistant_response = append_assistant_response
        self.metadata = metadata

    def poll_thread_execution(self, client:APIClient) -> dict:
        while True:
            try:
                thread_run_status = get_thread_execution_status(
                    client=client,
                    thread_execution_id=self.thread_execution_id
                ).status
            except Exception as e:
                print(e)
                raise Exception("failed to get thread execution status")
            if thread_run_status == "completed":
                break
            elif thread_run_status == "failed":
                raise Exception("Thread run failed")
            elif thread_run_status == "in_progress":
                print("thread run in progress")
                time.sleep(3)
            else:
                raise Exception(f"Unknown thread run status: {thread_run_status}")
        
        return get_thread_execution_response(
            client=client,
            thread_execution_id=self.thread_execution_id
        )
        
    def get_stop_reason(self, client:APIClient) -> str:
        execution_response = self.poll_thread_execution(client)
        return get_stop_reason_from_execution_response(execution_response)

class Tool:
    name: str
    description: str
    input_schema: dict

    def __init__(self, name:str, description:str, input_schema:dict):
        self.name = name
        self.description = description
        self.input_schema = input_schema

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "input_schema": self.input_schema
        }

def execute_messages(client:APIClient, thread_execution_param_id:str, messages:list[Message],system_prompt:str="", append_assistant_response:bool=True, metadata:dict={}) -> ThreadExecutionResponse:
    thread_id = "compext_thread_null"
    response = client.post(f"/thread/{thread_id}/execute", data={
            "thread_execution_param_id": thread_execution_param_id,
            "append_assistant_response": append_assistant_response,
            "thread_execution_system_prompt": system_prompt,
            "messages": [message.to_dict() for message in messages],
            "metadata": metadata,
        })

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to execute thread, status code: {status_code}, response: {data}")
        
    return ExecuteMessagesResponse(data["identifier"], thread_execution_param_id, messages, system_prompt, append_assistant_response, metadata)

class ExecuteMessagesWithToolsResponse(ExecuteMessagesResponse):
    tools: list[Tool]
    messages: list[Message]
    human_in_the_loop: bool
    human_intervention_handler: callable
    # new thread messages list keeps track of all the new messages added to the thread
    new_thread_messages: list[Message] = []
    def __init__(self, thread_execution_id:str, thread_execution_param_id:str, messages:list[Message], system_prompt:str, append_assistant_response:bool, metadata:dict, tools:list[str], human_in_the_loop:bool=False, human_intervention_handler:callable=None):
        super().__init__(thread_execution_id, thread_execution_param_id, messages, system_prompt, append_assistant_response, metadata)
        if human_in_the_loop:
            if human_intervention_handler is None:
                raise Exception("Human intervention handler is required when human_in_the_loop is True")
        self.tools = tools
        self.human_in_the_loop = human_in_the_loop
        self.human_intervention_handler = human_intervention_handler
        self.new_thread_messages = []
    def poll_until_completion(self, client:APIClient, execution_queue:queue.Queue=None) -> dict:
        while True:
            response = self.poll_thread_execution(client)
            if self.get_stop_reason(client) == "tool_calls":
                content_msg = response['response']['choices'][0]['message']['content']
                tool_calls = response['response']['choices'][0]['message']['tool_calls']
                self.messages.append(Message(
                    **response['response']['choices'][0]['message'],
                ))
                self.new_thread_messages.append(Message(
                    **response['response']['choices'][0]['message'],
                ))
                for tool_call in tool_calls:
                    tool_name = tool_call['function']['name']
                    tool_input = tool_call['function']['arguments']
                    tool_use_id = tool_call['id']

                    if tool_name == "json_tool_call":
                        self.messages.append(Message(
                            role="tool" ,
                            content=tool_call['function']['arguments'],
                            tool_call_id=tool_use_id
                        ))  
                        continue
                        
                    if execution_queue:
                        execution_queue.put({
                            "type": "tool_use",
                            "content": {
                                "tool_name": tool_name,
                                "tool_input": tool_input,
                                "tool_use_id": tool_use_id,
                                "content": content_msg
                            }
                        })
                    tool_input_dict = json.loads(tool_input)
                    print("tool return", tool_call)
                    print("tool input", tool_input_dict)
                    print("tool input type", type(tool_input_dict))
                    try:
                        if tool_name == "human_in_the_loop":
                            human_feedback = self.human_intervention_handler(**tool_input_dict)
                            tool_result = ToolResult(
                                result=human_feedback,
                                content=human_feedback,
                            )
                            self.new_thread_messages.append(Message(
                                role="tool" ,
                                content=tool_result.get_result(),
                                tool_call_id=tool_use_id
                            ))
                        else:
                            tool_obj = get_tool(tool_name)
                            tool_result = tool_obj(**tool_input_dict)
                            if tool_obj.append_tool_in_thread():
                                self.new_thread_messages.append(Message(
                                    role="tool" ,
                                    content=tool_result.get_content(),
                                    tool_call_id=tool_use_id
                                ))
                    except Exception as e:
                        print(f"Error executing tool {tool_name}: {e}")
                        raise Exception(f"Error executing tool {tool_name}: {e}")
        
                    # handle tool result
                    print(f"Tool {tool_name} returned: {tool_result.get_content()}")
                    if execution_queue:
                        execution_queue.put({
                            "type": "tool_result",
                            "content": {
                                "tool_use_id": tool_use_id,
                                "result": tool_result.get_content()
                            }
                        })
                    print("response assistant appending", response['response']['choices'][0]['message'])
                    self.messages.append(Message(
                        role="tool" ,
                        content=tool_result.get_result(),
                        tool_call_id=tool_use_id
                    ))
                # start a new execution with the new messages
                new_execution = execute_messages_with_tools(
                    client=client,
                    thread_execution_param_id=self.thread_execution_param_id,
                    messages=self.messages,
                    system_prompt=self.system_prompt,
                    append_assistant_response=self.append_assistant_response,
                    metadata=self.metadata,
                    tool_list=self.tools
                )
                self.thread_execution_id = new_execution.thread_execution_id
            else:
                break

        self.new_thread_messages.append(Message(
            role="assistant",
            content=response.get('content', '')
        ))
        return response
    
    def get_new_thread_messages(self) -> list[Message]:
        return self.new_thread_messages


def execute_messages_with_tools(client:APIClient, thread_execution_param_id:str, messages:list[Message],system_prompt:str="", append_assistant_response:bool=True, metadata:dict={}, tool_list:list[str]=[], human_in_the_loop:bool=False, human_intervention_handler:callable=None) -> ExecuteMessagesWithToolsResponse:
    if human_in_the_loop:
        tool_list.append("human_in_the_loop")
    thread_id = "compext_thread_null"
    response = client.post(f"/thread/{thread_id}/execute", data={
            "thread_execution_param_id": thread_execution_param_id,
            "append_assistant_response": append_assistant_response,
            "thread_execution_system_prompt": system_prompt,
            "messages": [message.to_dict() for message in messages],
            "metadata": metadata,
            "tools": [get_tool(tool).to_dict() for tool in tool_list]
        })
    
    status_code: int = response["status"]
    data: dict = response["data"]

    if status_code != 200:
        raise Exception(f"Failed to execute thread, status code: {status_code}, response: {data}")
    
    return ExecuteMessagesWithToolsResponse(data["identifier"], thread_execution_param_id, messages, system_prompt, append_assistant_response, metadata, tool_list, human_in_the_loop, human_intervention_handler)



from compextAI.api.api import APIClient
from datetime import datetime

class Message:
    message_id: str
    thread_id: str
    content: any
    role: str
    metadata: dict
    created_at: datetime
    updated_at: datetime
    tool_call_id: str
    tool_calls: any
    function_call: any

    def __init__(self,content:any, role:str, message_id:str='', thread_id:str='', metadata:dict={}, created_at:datetime=None, updated_at:datetime=None, **kwargs):
        self.tool_call_id = None
        self.tool_calls = None
        self.function_call = None
        self.message_id = message_id
        self.thread_id = thread_id
        self.content = content
        self.role = role
        self.metadata = metadata
        self.created_at = created_at
        self.updated_at = updated_at
        if "tool_call_id" in kwargs:
            self.tool_call_id = kwargs["tool_call_id"]
        if "tool_calls" in kwargs:
            self.tool_calls = kwargs["tool_calls"]
        if "function_call" in kwargs:
            self.function_call = kwargs["function_call"]

    def __str__(self):
        return f"Message(message_id={self.message_id}, thread_id={self.thread_id}, content={self.content}, role={self.role}, metadata={self.metadata}, tool_call_id={self.tool_call_id}, tool_calls={self.tool_calls}, function_call={self.function_call})"
    
    def to_dict(self) -> dict:
        return {
            "content": self.content,
            "role": self.role,
            "metadata": self.metadata,
            "message_id": self.message_id,
            "thread_id": self.thread_id,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "tool_call_id": self.tool_call_id if self.tool_call_id else None,
            "tool_calls": self.tool_calls if self.tool_calls else None,
            "function_call": self.function_call if self.function_call else None
        }

def get_message_object_from_dict(data:dict) -> Message:
    return Message(
        content=data["content"],
        role=data["role"],
        message_id=data["identifier"],
        thread_id=data["thread_id"],
        metadata=data["metadata"],
        created_at=data["created_at"],
        updated_at=data["updated_at"],
        tool_call_id=data["tool_call_id"] if "tool_call_id" in data else None,
        tool_calls=data["tool_calls"] if "tool_calls" in data else None,
        function_call=data["function_call"] if "function_call" in data else None
    )

def list_all(client:APIClient, thread_id:str) -> list[Message]:
    response = client.get(f"/message/thread/{thread_id}")

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to list messages, status code: {status_code}, response: {data}")
    
    return [get_message_object_from_dict(message) for message in data]

def retrieve(client:APIClient, message_id:str) -> Message:
    response = client.get(f"/message/{message_id}")

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to retrieve message, status code: {status_code}, response: {data}")
    
    return get_message_object_from_dict(data)

def create(client:APIClient, thread_id:str, messages:list[Message]) -> list[Message]:
    response = client.post(f"/message/thread/{thread_id}", data={"messages": [message.to_dict() for message in messages]})

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to create message, status code: {status_code}, response: {data}")
    
    return [get_message_object_from_dict(message) for message in data]

def update(client:APIClient, message_id:str, content:any, role:str, metadata:dict={}) -> Message:
    response = client.put(f"/message/{message_id}", data={"content": content, "role": role, "metadata": metadata})

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to update message, status code: {status_code}, response: {data}")
    
    return get_message_object_from_dict(data)

def delete(client:APIClient, message_id:str) -> bool:
    response = client.delete(f"/message/{message_id}")
    return response["status"] == 204



from compextAI.api.api import APIClient

class ThreadExecutionResponse:
    thread_execution_id: str
    thread_id: str

    def __init__(self, thread_execution_id:str, thread_id:str):
        self.thread_execution_id = thread_execution_id
        self.thread_id = thread_id

class Thread:
    thread_id: str
    title: str
    metadata: dict

    def __init__(self, thread_id:str, title:str, metadata:dict):
        self.thread_id = thread_id
        self.title = title
        self.metadata = metadata

    def __str__(self):
        return f"Thread(thread_id={self.thread_id}, title={self.title}, metadata={self.metadata})"
    
    def execute(self, client:APIClient, thread_exec_param_id: str, system_prompt:str="", append_assistant_response:bool=True, metadata:dict={}) -> ThreadExecutionResponse:
        response = client.post(f"/thread/{self.thread_id}/execute", data={
            "thread_execution_param_id": thread_exec_param_id,
            "append_assistant_response": append_assistant_response,
            "thread_execution_system_prompt": system_prompt,
            "metadata": metadata
        })

        status_code: int = response["status"]
        data: dict = response["data"]
        
        if status_code != 200:
            raise Exception(f"Failed to execute thread, status code: {status_code}, response: {data}")
        
        return ThreadExecutionResponse(data["identifier"], self.thread_id)


def get_thread_object_from_dict(data:dict) -> Thread:
    return Thread(data["identifier"], data["title"], data["metadata"])

def list(client:APIClient, project_name:str) -> list[Thread]:
    response = client.get(f"/thread/all/{project_name}")
    
    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to list threads, status code: {status_code}, response: {data}")
    
    threads = [get_thread_object_from_dict(thread) for thread in data]
    
    return threads

def retrieve(client:APIClient, thread_id:str) -> Thread:
    response = client.get(f"/thread/{thread_id}")
    
    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to retrieve thread, status code: {status_code}, response: {data}")
    
    return get_thread_object_from_dict(data)

def create(client:APIClient,project_name:str, title:str=None, metadata:dict={}) -> Thread:
    response = client.post(f"/thread", data={"project_name": project_name, "title": title, "metadata": metadata})

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to create thread, status code: {status_code}, response: {data}")
    
    return get_thread_object_from_dict(data)

def update(client:APIClient, thread_id:str, title:str=None, metadata:dict={}) -> Thread:
    response = client.put(f"/thread/{thread_id}", data={"title": title, "metadata": metadata})

    status_code: int = response["status"]
    data: dict = response["data"]
    
    if status_code != 200:
        raise Exception(f"Failed to update thread, status code: {status_code}, response: {data}")
    
    return get_thread_object_from_dict(data)

def delete(client:APIClient, thread_id:str) -> bool:
    response = client.delete(f"/thread/{thread_id}")
    return response["status"] == 204



def get_response_content_from_execution_response(execution_response: dict) -> str:
    return execution_response.get('content', '')

def get_response_object_from_execution_response(execution_response: dict) -> dict:
    return execution_response.get('response', {})

def get_stop_reason_from_response_object(response_object: dict) -> str:
    return response_object['choices'][0]['finish_reason']

def get_stop_reason_from_execution_response(execution_response: dict) -> str:
    return get_stop_reason_from_response_object(get_response_object_from_execution_response(execution_response))

def is_max_tokens_stop_reason(stop_reason: str) -> bool:
    return stop_reason == "length"
